;; Generated package description from corfu.el  -*- no-byte-compile: t -*-
(define-package "corfu" "0.38.0.20231127.210125" "COmpletion in Region FUnction" '((emacs "27.1") (compat "29.1.4.4")) :commit "9006b56e5db420cb82f86510f3d2c290036ade2a" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :keywords '("abbrev" "convenience" "matching" "completion" "wp") :url "https://github.com/minad/corfu")
